from poaupdater import pbaapi, uLogging, PEMVersion
import time

try:
    from hashlib import md5
except ImportError:
    from md5 import md5

ba_api = None


SUBSCRIPTION_STATUS = {10: "Ordered",
                       15: "Trial",
                       30: "Active",
                       40: "Graced",
                       50: "Expired",
                       60: "Terminated",
                       70: "Canceled",
                       80: "Administrative Hold",
                       85: "Credit Hold",
                       89: "Credit and  Administrative Hold"}

SUBSCRIPTION_SERVICE_STATUS = {10: "Not Provisioned",
                               20: "Provisioning",
                               30: "Stopped",
                               40: "Starting",
                               50: "Running",
                               60: "Stopping",
                               70: "Removing",
                               80: "Changing Plan",
                               90: "Removed"}


class OrderEx:
    def __init__(self, response):
        self.OrderID      = response[0]
        self.Total        = response[1]
        self.TaxTotal     = response[2]
        self.DiscTotal    = response[3]
        self.MerchTotal   = response[4]
        self.OrderNbr     = response[5]
        self.Descr        = response[6]
        self.CreationTime = response[7]
        self.DocID        = response[8]
        self.source = response


class _TypedMethodEx(pbaapi._TypedMethod):
    def __init__(self, api, server, name, username, password):
        version = PEMVersion.getPEMVersion().version
        if version[0] == 6:
            pbaapi._TypedMethod.__init__(self, api, server, name, username, password)  # 6.0.9 case
        else:
            pbaapi._TypedMethod.__init__(self, api, name)  # 7.x case
        self.resultTypes["PlacePlanPeriodSwitchOrder_API"] = OrderEx


class PBAAPIEx(pbaapi.PBAAPI):
    def __init__(self, autoCommit=True):
        pbaapi.PBAAPI.__init__(self, autoCommit)

    def __getattr__(self, name):
        return _TypedMethodEx(self, self.server, name, pbaapi._settings.username, pbaapi._settings.password)


def init_pba_api():
    pbaapi.initFromEnv()
    global ba_api
    ba_api = PBAAPIEx()


def get_subscription_details(sub_id, account_id):
    subs = ba_api.GetCustomerSubscriptionList_API(account_id, 0)
    sub = [s for s in subs if s[0] == sub_id][0]
    plan_id = sub[2]
    plan_period_id = sub[4]
    uLogging.debug("Subscription #%s: plan %s, period %s" % (sub_id, plan_id, plan_period_id))
    subscription_details = ba_api.SubscriptionDetailsGet_API(sub_id)
    return plan_id, plan_period_id, subscription_details.Status, subscription_details.ServStatus


def is_subscription_active(sub_status, sub_serv_status):
    return sub_status == 30 and sub_serv_status == 50


def is_subscription_changing_plan(sub_status, sub_serv_status):
    return sub_status == 30 and sub_serv_status == 80


def validate_plans_map(settings):
    uLogging.info("Validate Service Plans")
    for source_plan_id in settings.get_source_plans():
        target_plan_id = settings.get_target_plan(source_plan_id)
        check_service_plan_presence(source_plan_id)
        check_service_plan_presence(target_plan_id)
        for source_period_id in settings.get_source_periods(source_plan_id):
            check_subscription_period(source_plan_id, source_period_id)
            check_subscription_period(target_plan_id, settings.get_target_period(source_plan_id, source_period_id))


def check_service_plan_presence(plan_id):
    uLogging.info("Check presence of Service Plan # %s" % plan_id)
    plan = ba_api.PlanDetailsGet_API(plan_id)
    uLogging.info("Found Service Plan #%s '%s'" % (plan[0], plan[1]))


def check_subscription_period(plan_id, subscription_period_id):
    subscription_period = ba_api.PlanPeriodDetailsGet_API(subscription_period_id)
    is_period_owned_by_target_plan = plan_id == subscription_period[1]
    if not is_period_owned_by_target_plan:
        raise Exception("Subscription Period #%s is not owned by Service Plan #%s. Please correct settings.py."
                        % (subscription_period_id, plan_id))


def create_switch_order(sub_id, source_plan_id, target_plan_id, target_period_id):
    upgrade_start_type = 10  # UpgradeStartType = 10 means "from the start date of current subscription period"
    apply_promo = 0
    uLogging.info("Switch subcription #%s from Plan #%s to Plan #%s, subscription period #%s"
                  % (sub_id, source_plan_id, target_plan_id, target_period_id))
    order = ba_api.PlacePlanPeriodSwitchOrder_API(sub_id, target_plan_id, target_period_id, upgrade_start_type, apply_promo)
    uLogging.debug("Order %s has been created for subscription %s" % (order.OrderID, sub_id))
    return order.OrderID


def open_switch_order(order_id):
    order = ba_api.GetOrder_API(order_id)
    order_number = str(order[1]).strip()
    status = str(order[4])
    creation_time = str(int(order[6]))
    amount = "%.2f" % order[8]
    description = str(order[12])
    currency = str(order[17]).strip()

    if status == 'NW':
        uLogging.debug("Open order %s" % order_id)
        signature_parts = [str(order_id), order_number, creation_time, currency, " ", amount, description]
        signature = md5("".join(signature_parts).encode('utf-8')).hexdigest()
        try:
            ba_api.OrderStatusChange_API(order_id, "OP", signature)
        except Exception, e:
            order = ba_api.GetOrder_API(order_id)
            status = str(order[4])
            if status == 'NW':
                uLogging.warn("Failed to open order %s" % order_id)
                raise e
            else:
                uLogging.debug("Order %s is already opened by system." % order_id)


def get_order_info(order_id):
    return ba_api.GetOrder_API(order_id)


def is_order_completed(order_info):
    status = str(order_info[4])
    return status == 'CP'


def is_order_waiting_for_paiment(order_info):
    status = str(order_info[4])
    return status in ['NP', 'WP', 'WZ']


def is_order_canceled(order_info):
    status = str(order_info[4])
    return status == 'CL'


def is_order_failed(order_info):
    status = str(order_info[4])
    return status in ['PN', 'PF']


def is_order_waiting_long_running_operation(order_info):
    status = str(order_info[4])
    return status == 'LO'


def is_order_in_known_status(order_id):
    '''
    check if order is in one of known states
    LongRunning - means migration tasks are created
    Failed, Canceled, Waiting For Payment - smth goes wrong, need continue other subscriptions processing
    '''
    order_info = get_order_info(order_id)
    return is_order_failed(order_info) \
           or is_order_waiting_long_running_operation(order_info) \
           or is_order_waiting_for_paiment(order_info) \
           or is_order_canceled(order_info) \
           or is_order_completed(order_info)
