import os
import sys
import re
from datetime import datetime
import optparse
import time
import billing
from poaupdater import openapi, uConfig, uLogging, uSysDB
from settings import MigrationSettings


SECONDS_15 = 15
MINUTES_1 = 60

MISSED_PARAMETER = 1

# migration statuses
SCHEDULING = 10
SCHEDULED = 20
FAILED = 30
WAITING_PAYMENT = 40
CANCELED = 50
COMPLETED = 60


class MigrationScheduler:
    '''
    Schedules subscriptions migration.
    Number of simultaneously scheduling subscriptions is limited by 'concurrency_limit' parameter.
    If 'concurrency_limit' value is reached, script waits 15 seconds before next status update.
    If script detects non-used slots, it schedules additional subscriptions migration until 'concurrency_limit' value
    is reached (or until all subscriptions has been processed).
    After subscription migration launch, script switches target Mailbox server with help of "Ready to Provide"
    parameter in order to distribute mailboxes among servers listed in 'target_mbx_servers' parameter.
    Subscription is considered as 'scheduling'
        - from the moment of Change Order creation and
        - to the moment of all migration tasks has been processed (except "Waiting completion" and "Finalize" tasks)
    If Change Order fails or any migration task fails, subscription migration is considered as failed.
    Number of failures is limited by 'failures_limit' parameter. If limit is reached, script stops working.
    If subscription state doesn't allow to schedule subscription, script skips such subscription.
    '''

    def __init__(self, subscriptions, settings):
        self.to_migrate = subscriptions
        self.status = MigrationStatus()
        self.server_switcher = ServerSwitcher(settings.servers)
        self.settings = settings

    def schedule_migration(self):
        while not self._is_scheduling_completed():
            self.status.update()
            while self._is_subscription_scheduling_allowed() and not self._is_failures_limit_reached():
                sub_id = self.to_migrate[0]
                db_con = DatabaseAccessor()
                account = db_con.get_account(sub_id)
                if account:
                    self._init_subscription_migration(sub_id, account, db_con)
                else:
                    self.status.subscription_not_found(sub_id)
                self.to_migrate.remove(sub_id)
            self.status.brief_report(self.to_migrate)
            if self.status.scheduling:
                uLogging.info("Waiting for a 15 sec...")
                time.sleep(SECONDS_15)
        if self._is_failures_limit_reached():
            uLogging.err("Number of failures reached defined limit %s" % self.settings.failures_limit)

    def report_status(self):
        self.status.report(self.to_migrate)

    def reset_servers(self):
        self.server_switcher.reset_to_initial_state()

    def _is_scheduling_completed(self):
        '''
        Subscriptions migration scheduling is completed when
            - list of 'scheduling' subscriptions is empty AND
            - all subscriptions has been processed OR failures limit reached
        '''
        return not self.status.scheduling and (not self.to_migrate or self._is_failures_limit_reached())

    def _is_failures_limit_reached(self):
        '''
        Failures limit is reached when
            - 'failures_limit' value is not negative AND
            - number of failures has reached or exceeded 'failures_limit' value
        :return:
        '''
        return 0 <= self.settings.failures_limit <= len(self.status.failed)

    def _is_subscription_scheduling_allowed(self):
        '''
        It's allowed to launch one more subscription migration if
            - 'concurrency_limit' is not reached yet AND
            - not all subscriptions has been processed yet
        '''
        return len(self.status.scheduling) < self.settings.concurrency_limit and self.to_migrate

    def _init_subscription_migration(self, sub_id, account, db_con):
        '''
        Launches migration of particular subscription.
            - checks subscription status
            - initiates switching Plan for subscription if possible
        '''
        source_plan_id, source_period_id, sub_status, sub_serv_status = billing.get_subscription_details(sub_id, account)

        if not self.settings.is_source_plan(source_plan_id):
            self.status.incorrect_source_plan_skipped(sub_id, source_plan_id, self.settings.is_target_plan(source_plan_id))
        elif billing.is_subscription_changing_plan(sub_status, sub_serv_status):
            self.status.change_order_exists(sub_id)
        elif billing.is_subscription_active(sub_status, sub_serv_status):
            # In 6.0.9 Billing allows to create new Change Order for subscription while there is failed Change Order
            # Moreover, new order somehow becomes Completed w/o tasks scheduling
            # So, we need prevent new order creation if there are existing migration tasks in OA
            tasks = db_con.get_migration_edge_tasks(sub_id)
            if tasks:
                self.status.migration_tasks_exist(sub_id, source_plan_id, tasks)
            else:
                subscription_related_tasks = db_con.get_tasks(sub_id)
                if subscription_related_tasks:
                    self.status.tasks_exist_before_migration(sub_id, subscription_related_tasks)
                else:
                    self._switch_plan(sub_id, source_plan_id, source_period_id, db_con)
        else:
            self.status.incorrect_status_skipped(sub_id, sub_status, sub_serv_status)

    def _switch_plan(self, sub_id, current_plan, current_period, db_con):
        '''
        Creates Change Order for subscription.
        If order creation fails, updates subscription status accordingly.
        If order created successfully, waits until order turns into one of "known" states
            - Long Running Operation - means that task "Migration: Prepare for Exchange organization" created in OA
            - Failed
            - Canceled
            - Waiting for Payment
        If order is in "known" state, waits for completion of task "Migration: Prepare for Exchange organization",
        which means that all migration tasks has been created and switches target Mailbox server.

        :param sub_id:
        :param current_plan: Service Plan the subscription is based on
        :param current_period: Subscription Period the subscription is based on
        '''
        try:
            order_id = billing.create_switch_order(sub_id,
                                                   current_plan,
                                                   self.settings.get_target_plan(current_plan),
                                                   self.settings.get_target_period(current_plan, current_period))
            billing.open_switch_order(order_id)
            self._wait_for_tasks_creation(order_id, sub_id, db_con)
        except Exception, e:
            self.status.change_order_creation_failed(sub_id, str(e))
        else:
            self.server_switcher.switch_server()
            self.status.scheduling_started(sub_id, order_id)

    def _wait_for_tasks_creation(self, order_id, sub_id, db_con):
        counter = 0
        order_in_known_status = False
        while counter < MINUTES_1:
            preparing_task_succeeded = db_con.is_org_migration_preparing_task_succeeded(sub_id)
            order_in_known_status = billing.is_order_in_known_status(order_id)
            if order_in_known_status or preparing_task_succeeded:
                break
            preparing_task_failed = db_con.is_org_migration_preparing_task_failed(sub_id)
            if preparing_task_failed:
                break
            else:
                counter += 1
                time.sleep(1)

        if not order_in_known_status:
            self.status.order_stuck_in_unexpected_state(sub_id)  # add stuck order to report and continue work
            preparing_task_absent_or_failed = db_con.is_org_migration_preparing_task_absent_or_failed(sub_id)
            preparing_task_succeeded = db_con.is_org_migration_preparing_task_succeeded(sub_id)
            if preparing_task_absent_or_failed and not preparing_task_succeeded:
                raise Exception("Order %s has stuck in unexpected state; migration tasks are not created" % order_id)
            else:
                uLogging.warn("Order %s has stuck in unexpected state; migration tasks are being created" % order_id)
        db_con.wait_for_org_migration_preparing_task_absence(sub_id)


class MigrationStatus:
    '''
    Class keeps overall migration status for all subscriptions
    Each subscription is placed to one of available lists (see in ctor)
    Number of subscriptions in 'scheduling' list is limited by settings parameter 'concurrency_limit'
    When status of 'scheduling' subscription has changed, it's moved to appropriate list
    All lists except 'scheduling' represent final subscription state: subscription is never moved from these lists
    '''

    def __init__(self):
        self.scheduling = []                            # list of SubscriptionMigrationData being scheduling
        self.scheduled = []                             # list of scheduled SubscriptionMigrationData
        self.failed = []                                # list of failed SubscriptionMigrationData
        self.waiting_payment = []                       # list of waiting for payment SubscriptionMigrationData
        self.canceled = []                              # list SubscriptionMigrationData with canceled orders
        self.skipped_bad_status = []                    # list of subscription IDs skipped due to non-applicable subscription status
        self.skipped_already_on_target = []             # list of subscription IDs skipped as already migrated
        self.skipped_wrong_plan_configuration = []      # list of subscription IDs skipped due to incorrect source/target plan settings
        self.order_stuck = []                           # list of subscription IDs having orders stuck in unexpected state

    def update(self):
        '''
        Updates actual status for subscriptions listed in 'scheduling' list
        '''
        db_con = DatabaseAccessor()
        queue_map = self._get_queue_map()
        scheduling_completed = []
        for migration_data in self.scheduling:
            migration_data.update_status(db_con)
            if migration_data.status != SCHEDULING:
                queue_map[migration_data.status].append(migration_data)
                scheduling_completed.append(migration_data)
        self.scheduling = [md for md in self.scheduling if md not in scheduling_completed]

    def incorrect_source_plan_skipped(self, sub_id, subscription_plan_id, is_target_plan):
        '''
        Places subscription with incorrect source plan to appropriate list
        '''
        if is_target_plan:
            uLogging.info("Subscription #%s is already based on target plan %s, nothing to do." % (sub_id, subscription_plan_id))
            self.skipped_already_on_target.append(sub_id)
        else:
            uLogging.warn("Cannot schedule subscription migration: "
                          "subscription #%s is based on plan %s, which is neither source plan nor target plan." % (sub_id, subscription_plan_id))
            self.skipped_wrong_plan_configuration.append(sub_id)

    def incorrect_status_skipped(self, sub_id, sub_status, sub_serv_status):
        '''
        Places subscription having incorrect status to appropriate list
        '''
        uLogging.warn("Cannot switch plan for subscription #%s (status: %s, service status: %s). "
                      "It's only allowed to migrate Active and Running subscriptions."
                      % (sub_id,
                         billing.SUBSCRIPTION_STATUS[sub_status],
                         billing.SUBSCRIPTION_SERVICE_STATUS[sub_serv_status]))
        self.skipped_bad_status.append(sub_id)

    def subscription_not_found(self, sub_id):
        '''
        Places invalid subscription having no account to appropriate list
        '''
        uLogging.err("Account owning subscription %s is not found. Skip subscription migration." % sub_id)
        self.skipped_bad_status.append(sub_id)

    def change_order_exists(self, sub_id):
        '''
        Places subscription having status 'Changing Plan' to appropriate list
        '''
        uLogging.info("Subscription #%s is in state Changing Plan. No need to create new Change Order" % sub_id)
        self.scheduling.append(NoOrderSubscriptionMigrationData(sub_id))

    def order_stuck_in_unexpected_state(self, sub_id):
        '''
        Adds subscription having order which has stuck in unexpected state status to appropriate list
        '''
        uLogging.info("Subscription #%s has order stuck in unexpected state. Please, re-check order state." % sub_id)
        self.order_stuck.append(sub_id)

    def change_order_creation_failed(self, sub_id, error_message):
        '''
        Places subscription to appropriate list in case of Change Order creation failure
        '''
        uLogging.debug(error_message)
        if "This order prohibits placing an order of the same order type" in error_message:
            uLogging.debug("Subscription #%s already has non-completed Change Order." % sub_id )
            self.scheduling.append(NoOrderSubscriptionMigrationData(sub_id))
        else:
            uLogging.warn("Change Order creation failed for subscription #%s." % sub_id )
            self.failed.append(NoOrderSubscriptionMigrationData(sub_id, status=FAILED))

    def scheduling_started(self, sub_id, order_id):
        '''
        Places to 'scheduling' list subscription with known Change Order
        '''
        self.scheduling.append(OrderedSubscriptionMigrationData(sub_id, order_id=order_id))

    def migration_tasks_exist(self, sub_id, source_plan_id, tasks):
        '''
        Places to 'scheduling' list subscription without known Change Order but having migration tasks in OA
        '''
        uLogging.info("Subscription #%s based on source plan %s is Active Running, but there are migration tasks in OA. "
                      "No need to create new Switch Order." % (sub_id, source_plan_id))
        uLogging.debug("Tasks found in OA: \n%s" % "\n".join(["%s  %s  %s" % (task.id, task.name, task.status) for task in tasks]))
        self.scheduling.append(NoOrderSubscriptionMigrationData(sub_id))

    def tasks_exist_before_migration(self, sub_id, tasks):
        '''
        Places to 'skipped_bad_status' list subscription having non-processed non-migration tasks in OA
        '''
        uLogging.info("Migration of subscription #%s is skipped: there are tasks in OA related to the subscription. "
                      "These tasks should be processed before migration start." % sub_id)
        uLogging.debug("Tasks found in OA: \n%s" % "\n".join(["%s  %s  %s" % (task.id, task.name, task.status) for task in tasks]))
        self.skipped_bad_status.append(sub_id)

    def _get_queue_map(self):
        return {
            SCHEDULING: self.scheduling,
            SCHEDULED: self.scheduled,
            FAILED: self.failed,
            WAITING_PAYMENT: self.waiting_payment,
            CANCELED: self.canceled,
        }

    def _get_progress(self, not_started):
        '''
        Calculates current progress
        :param not_started: list of subscriptions, which processing wasn't started yet
        :return: percent, processed, total
                 percent - percent of processed subscriptions
                 processed - number of processed subscriptions
                 total - total number of subscriptions
        '''
        processed_subs_number = len(self.scheduled) + len(self.failed) + len(self.waiting_payment) \
                                + len(self.canceled) + len(self.skipped_bad_status) \
                                + len(self.skipped_already_on_target) + len(self.skipped_wrong_plan_configuration)
        subs_number = processed_subs_number + len(not_started) + len(self.scheduling)
        return "%s%%" % int(float(processed_subs_number)/subs_number * 100), processed_subs_number, subs_number

    def report(self, not_started):
        '''
        Prints final migration report.
        :param not_started: list of subscriptions, which processing wasn't started yet
        '''
        percent, processed, total = self._get_progress(not_started)
        report_params = {'percent': percent,
                         'processed_nmb': processed,
                         'total_nmb': total,
                         'on_target_nmb': len(self.skipped_already_on_target),
                         'scheduled_nmb': len(self.scheduled),
                         'bad_status_nmb': len(self.skipped_bad_status),
                         'bad_status_subs': self.skipped_bad_status,
                         'wrong_plan_nmb': len(self.skipped_wrong_plan_configuration),
                         'wrong_plan_subs': self.skipped_wrong_plan_configuration,
                         'failed_nmb': len(self.failed),
                         'failed_subs': [s.sub_id for s in self.failed],
                         'waiting_payment_nmb': len(self.waiting_payment),
                         'waiting_payment_subs': [s.sub_id for s in self.waiting_payment],
                         'scheduling_nmb': len(self.scheduling),
                         'scheduling_subs': [s.sub_id for s in self.scheduling],
                         'stuck_orders': self.order_stuck,
                         'stuck_orders_nmb': len(self.order_stuck),
                         }
        if self.skipped_bad_status:
            uLogging.info("%(bad_status_nmb)s subscriptions have a status that prevents to switch Plan. "
                          "Check subscription status in Billing, subscription-related tasks in OA task manager, "
                          "subscription existence in OA. "
                          "See log for details. Affected subscriptions: %(bad_status_subs)s" % report_params)
        if self.skipped_wrong_plan_configuration:
            uLogging.info("%(wrong_plan_nmb)s subscriptions were skipped (could not identify target Plan). "
                          "See log for details. Affected subscriptions: %(wrong_plan_subs)s" % report_params)
        if self.failed:
            uLogging.info("%(failed_nmb)s subscriptions migration failed. "
                          "Check order status in Billing, subscription-related tasks in OA task manager. "
                          "See log for details. Affected subscriptions: %(failed_subs)s" % report_params)
        if self.waiting_payment:
            uLogging.info("%(waiting_payment_nmb)s subscriptions are waiting for payment. "
                          "Check order status in Billing. "
                          "Affected subscriptions: %(waiting_payment_subs)s" % report_params)
        if self.scheduling:
            uLogging.info("%(scheduling_nmb)s subscriptions are being scheduled. "
                          "Check order status in Billing, subscription-related tasks in OA task manager. "
                          "Affected subscriptions: %(scheduling_subs)s" % report_params)
        if self.order_stuck:
            uLogging.info("%(stuck_orders_nmb)s subscriptions has orders stuck in unexpected state. "
                          "Check order status in Billing. "
                          "Affected subscriptions: %(stuck_orders)s" % report_params)

        result = """
                
        %(percent)s processed (%(processed_nmb)s of %(total_nmb)s subscriptions)
        ============================================================================================
        %(on_target_nmb)s subscriptions are already on target Plan.
        %(scheduled_nmb)s subscriptions are scheduled for migration.        
        %(bad_status_nmb)s subscriptions have a status that prevents to switch Plan.
        %(wrong_plan_nmb)s subscriptions were skipped (could not identify target Plan).
        %(failed_nmb)s subscriptions migration failed.        
        %(waiting_payment_nmb)s subscriptions are waiting for payment.
        %(scheduling_nmb)s subscriptions are being scheduled.
        %(stuck_orders_nmb)s subscriptions has orders with unexpected state.
        ============================================================================================
        """ % report_params
        uLogging.info(result)
        uLogging.debug("Non-processed subscriptions: %s" % not_started)

    def brief_report(self, not_started):
        '''
        Prints brief intermediate migration report.
        :param not_started: list of subscriptions, which processing wasn't started yet
        '''
        percent, processed, total = self._get_progress(not_started)
        uLogging.info("""
        
        %(percent)s processed (%(processed_nmb)s of %(total_nmb)s subscriptions)
        Not started yet %(not_started_nmb)s. Scheduling %(scheduling_nmb)s. Waiting for payment %(waiting_payment_nmb)s. \
Skipped %(skipped_nmb)s. Failed %(failed_nmb)s. Scheduled %(scheduled_nmb)s. Already on target plan %(on_target_nmb)s.
        Orders in unexpected state: %(stuck_orders_nmb)s
        """ % {'percent': percent,
               'processed_nmb': processed,
               'total_nmb': total,
               'not_started_nmb': len(not_started),
               'on_target_nmb': len(self.skipped_already_on_target),
               'scheduled_nmb': len(self.scheduled),
               'skipped_nmb': len(self.skipped_bad_status + self.skipped_wrong_plan_configuration),
               'failed_nmb': len(self.failed),
               'waiting_payment_nmb': len(self.waiting_payment),
               'scheduling_nmb': len(self.scheduling),
               'stuck_orders_nmb': len(self.order_stuck),
               })


class SubscriptionMigrationData:
    '''
    Class stores migration data and status for particular subscription
    It's abstract and has two children handling status updating:
        - one for case when ID of Change Order is known (order was created by script)
        - one for case when ID of Change Order is NOT known (order was created before script launch)
    '''
    def __init__(self, sub_id, status=SCHEDULING):
        self.sub_id = sub_id
        self.status = status

    def update_status(self, db_con):
        pass

    def _set_canceled(self):
        uLogging.warn("Change Order is canceled (subscription #%s)." % self.sub_id)
        self.status = CANCELED

    def _set_waiting_payment(self):
        uLogging.warn("Subscription #%s is waiting for payment." % self.sub_id)
        self.status = WAITING_PAYMENT

    def _set_failed(self, db_con):
        uLogging.warn("Subscription #%s migration has failed." % self.sub_id)
        self._log_failed_tasks(db_con)
        self.status = FAILED

    def _set_scheduled(self):
        uLogging.info("Subscription #%s migration has been scheduled." % self.sub_id)
        self.status = SCHEDULED

    def _log_failed_tasks(self, db_con):
        failed_tasks = db_con.get_failed_tasks(self.sub_id)
        if failed_tasks:
            uLogging.debug("Failed OA tasks related to subscription #%s:" % self.sub_id)
            for task in failed_tasks:
                uLogging.debug("%15s  %s" % (task.id, task.name))


class OrderedSubscriptionMigrationData(SubscriptionMigrationData):
    '''
    handles subscription migration status updating when Change Order ID is known (order was created by script)
    '''

    def __init__(self, sub_id, order_id, status=SCHEDULING):
        SubscriptionMigrationData.__init__(self, sub_id, status=status)
        self.order_id = order_id

    def update_status(self, db_con):
        order_info = billing.get_order_info(self.order_id)
        if billing.is_order_canceled(order_info):
            self._set_canceled()
        elif billing.is_order_waiting_for_paiment(order_info):
            self._set_waiting_payment()
        elif billing.is_order_failed(order_info) or db_con.get_failed_tasks(self.sub_id):
            self._set_failed(db_con)
        elif not db_con.get_pre_finalization_migration_tasks(self.sub_id):
            self._set_scheduled()


class NoOrderSubscriptionMigrationData(SubscriptionMigrationData):
    '''
    handles subscription migration status updating when Change Order ID is NOT known, i.e.
        - order existed before script launch
        - order is not completed yet (subscription is not on target plan yet)
    '''

    def __init__(self, sub_id, status=SCHEDULING):
        SubscriptionMigrationData.__init__(self, sub_id, status=status)

    def update_status(self, db_con):
        if not db_con.get_migration_edge_tasks(self.sub_id):
            # migration tasks absence means one of following: order either isn't opened yet or isn't paid yet
            # we cannot distinguish these cases, so report both as waiting for payment to attract attention to order
            self._set_waiting_payment()
        elif db_con.get_failed_tasks(self.sub_id):
            self._set_failed(db_con)
        elif not db_con.get_pre_finalization_migration_tasks(self.sub_id):
            self._set_scheduled()


class ServerSwitcher:
    def __init__(self, servers):
        self.servers = servers
        self.server_index = 0
        self.servers_number = len(self.servers)
        self.initial_state = {}
        for host_id in servers:
            server_data = oa_api.pem.getHost(host_id=host_id)
            self.initial_state[host_id] = server_data['ready_to_provide']
            if server_data['ready_to_provide'] and host_id != self.servers[0]:
                oa_api.pem.setHostReadyToProvide(host_id=host_id, ready_to_provide=False)
        oa_api.pem.setHostReadyToProvide(host_id=servers[0], ready_to_provide=True)
        uLogging.debug("Initial servers state: %s" % self.initial_state)

    def switch_server(self):
        current_mbx_id = self.servers[self.server_index]
        self.server_index = (self.server_index + 1) % self.servers_number
        next_mbx_id = self.servers[self.server_index]
        if current_mbx_id != next_mbx_id:
            uLogging.debug("Mark host #%s as ready to provide." % next_mbx_id)
            oa_api.pem.setHostReadyToProvide(host_id=next_mbx_id, ready_to_provide=True)
            oa_api.pem.setHostReadyToProvide(host_id=current_mbx_id, ready_to_provide=False)

    def reset_to_initial_state(self):
        uLogging.debug("Set hosts ready to provide settings to initial state.")
        for host_id in self.servers:
            oa_api.pem.setHostReadyToProvide(host_id=host_id, ready_to_provide=self.initial_state[host_id])


class PoaTask:
    def __init__(self, task_id, task_name, status):
        self.id = task_id
        self.name = task_name
        self.status = status


class DatabaseAccessor:
    def __init__(self):
        self.con = uSysDB.connect()

    def get_account(self, sub_id):
        sql = "SELECT owner_id FROM subscriptions WHERE sub_id = %s"
        cur = self.con.cursor()
        cur.execute(sql, sub_id)
        row = cur.fetchone()
        # return row[0] if row else None  - is not supported in python 2.4, Centos5
        if row:
            return row[0]
        return None


    def get_tasks(self, sub_id):
        sql = "SELECT task_id, name, status FROM tm_tasks WHERE subscription_id = %s"
        cur = self.con.cursor()
        cur.execute(sql, sub_id)
        rows = cur.fetchall()
        return [PoaTask(row[0], row[1], row[2]) for row in rows]

    def get_failed_tasks(self, sub_id):
        sql = "SELECT task_id, name, status FROM tm_tasks WHERE status = 'f' AND subscription_id = %s"
        cur = self.con.cursor()
        cur.execute(sql, sub_id)
        rows = cur.fetchall()
        return [PoaTask(row[0], row[1], row[2]) for row in rows]

    def wait_for_org_migration_preparing_task_absence(self, sub_id):
        '''
        waits until task creating real migration tasks is absent in task manager.
        target MBX server switching is safe only after 'Migration: Prepare for Exchange organization %%' task completion
        method prerequisite: Change order is in one of "known" states: long running, failed, waiting payment, canceled

        task is either created (order is LO) or failed (order Failed) or absent (order processed or waiting for payment)
        all these cases are valid and will be correctly handled while updating report
        note: we cannot rely to status in tm_logs since it's possible that task was not created
        '''
        counter = 0
        while True:
            counter += 1
            if self.is_org_migration_preparing_task_absent_or_failed(sub_id):  # (task is executed or wasn't created) or (task failed)
                break
            if counter > SECONDS_15:
                uLogging.info("Task 'Migration: Prepare for Exchange organization' for subscription %s is not executed yet, waiting..." % (sub_id))
                counter = 0
            time.sleep(1)

    def is_org_migration_preparing_task_absent_or_failed(self, sub_id):
        '''
        checks presence and status of task 'Migration: Prepare for Exchange organization %%' in Task Manager
        returns True if task is absent or has status Failed
        '''
        task = self.get_org_migration_preparing_task(sub_id)
        if task:
            return task.status == 'f'
        else:
            return True

    def is_org_migration_preparing_task_failed(self, sub_id):
        '''
        checks if there is failed task 'Migration: Prepare for Exchange organization %%' in Task Manager
        returns True if failed task is found
        '''
        task = self.get_org_migration_preparing_task(sub_id)
        if task:
            return task.status == 'f'
        else:
            return False

    def get_org_migration_preparing_task(self, sub_id):
        '''
        checks presence and status of task 'Migration: Prepare for Exchange organization %%' in Task Manager
        returns PoaTask if task is found or None otherwise
        '''
        sql = "SELECT task_id, name, status FROM tm_tasks WHERE name like 'Migration: Prepare for Exchange organization %%' AND subscription_id = %s"
        cur = self.con.cursor()
        cur.execute(sql, sub_id)
        row = cur.fetchone()
        if row:
            uLogging.debug("Task #%s '%s' is not executed yet, current status is '%s'..." % (row[0], row[1], row[2]))
            return PoaTask(row[0], row[1], row[2])
        else:
            return None

    def is_org_migration_preparing_task_succeeded(self, sub_id):
        '''
        checks presence and status of task 'Migration: Prepare for Exchange organization %%' in Task Log
        returns True if task is present in Task Log and has status Success
        '''
        sql = "SELECT task_id, name, exec_status FROM tm_logs WHERE name like 'Migration: Prepare for Exchange organization %%' AND subscription_id = %s"
        cur = self.con.cursor()
        cur.execute(sql, sub_id)
        rows = cur.fetchall()
        for row in rows:
            if row[2] == 's':
                return True
        return False

    def get_migration_edge_tasks(self, sub_id):
        '''
        checks if task manager contains migration tasks related to subscription
        task names may differ in 6.0 and 7.x, so both variants are present in the list
        only "first" and "last" migration tasks are taken into account; all intermediate are ignored
        '''
        sql = "SELECT task_id, name, status " \
              "FROM tm_tasks " \
              "WHERE (" \
              "name like 'Migration: Prepare for Exchange organization %%' " \
              "OR name like 'Waiting completion of move request for mailbox %%' " \
              "OR name like 'Waiting for the completion of the move request of the mailbox %%' " \
              "OR name like 'Finalize the mailbox %%' " \
              "OR name like 'Finalize the migration of the mailbox %%'" \
              ") " \
              "AND subscription_id = %s"
        cur = self.con.cursor()
        cur.execute(sql, sub_id)
        rows = cur.fetchall()
        return [PoaTask(row[0], row[1], row[2]) for row in rows]

    def get_pre_finalization_migration_tasks(self, sub_id):
        '''
        checks if task manager contains migration tasks related to subscription
        task names may differ in 6.0 and 7.x, so both variants are present in the list
        all migration tasks are taken into account except last tasks ("Waiting for completion" and "Finalize")
        all these tasks should be executed before subscription is considered as 'scheduled'
        '''
        sql = "SELECT task_id, name, status " \
              "FROM tm_tasks " \
              "WHERE (" \
              "name like 'Migration: Prepare for Exchange organization %%' " \
              "OR name like 'Migrate Address List to Exchange %%' " \
              "OR name like 'Migrate Global Address List to Exchange %%' " \
              "OR name like 'Migrate contacts to Exchange %%' " \
              "OR name like 'Migrate groups to Exchange %%' " \
              "OR name like 'Migrate Offline Address Book to Exchange %%' " \
              "OR name like 'Migrate the mailbox %%' " \
              ") " \
              "AND subscription_id = %s"
        cur = self.con.cursor()
        cur.execute(sql, sub_id)
        rows = cur.fetchall()
        return [PoaTask(row[0], row[1], row[2]) for row in rows]


def get_subscriptions_to_migrate(filenames):
    subscriptions = set()
    pattern = re.compile("^\s*\d+\s*$")
    for filename in filenames:
        uLogging.info("Read subscriptions list from file '%s'." % filename)
        f = open(filename)  # with open(filename) as f  is not supported in python 2.4, Centos5
        content = f.readlines()
        f.close()
        for sub_id in content:
            if pattern.match(sub_id):
                subscriptions.add(int(sub_id))
            else:
                uLogging.debug("Failed to parse subscription id from value '%s'." % sub_id.replace('\n', ''))
    return list(subscriptions)


def init_logger():
    logdir = os.path.dirname(os.path.abspath(__file__))
    logfile = '%s/schedule_migration.%s.log' % (logdir, datetime.now().strftime('%Y-%m-%d-%H-%M-%S'))

    config = uConfig.Config(
        rootpath='/usr/local/pem',  # to prevent import of uLinux which imports yum module which may be absent on MN
        verbose=True  # verbose=True is required in 6.0.9, can be skipped in 7.4
    )
    config.log_file = logfile
    uLogging.init(config)

    class FakeStream:
        def write(self, *args, **kwargs):
            pass
        def flush(self, *args, **kwargs):
            pass
    uLogging.debug = uLogging.log_func(FakeStream(), uLogging.DEBUG)  # print debug logs to file only

    return logfile


def init_options():
    usage = """
         %prog -s|--settings <settings file> <subscriptions list file>...
Examples: 
         %prog -s settings.json subscriptions.list
         %prog -s settings.json subscriptions.list more_subscriptions.list
    """
    parser = optparse.OptionParser(usage)
    parser.add_option("-s", "--settings", dest="settings", help="Path to migration settings json file")
    (options, args) = parser.parse_args()

    if options.settings is None:
        uLogging.err("Migration settings file is not specified. Please, correct input parameters.")
        parser.print_help()
        sys.exit(MISSED_PARAMETER)
    if not args:
        uLogging.err("File containing list of subscriptions to be scheduled is not specified. Please, correct input parameters.")
        parser.print_help()
        sys.exit(MISSED_PARAMETER)
    return options, args


def main():
    options, filenames = init_options()
    uLogging.info("Launched with parameters %s" % options)
    settings = MigrationSettings(options.settings)
    billing.validate_plans_map(settings)

    subscriptions = get_subscriptions_to_migrate(filenames)
    scheduler = MigrationScheduler(subscriptions, settings)
    try:
        scheduler.schedule_migration()
    finally:
        scheduler.reset_servers()
        scheduler.report_status()
        uLogging.info("Log file is located in: %s" % log_file)


log_file = init_logger()
billing.init_pba_api()
oa_api = openapi.OpenAPI()
if __name__ == "__main__":
    main()

