from poaupdater import uLogging
try:
    import json
except ImportError:
    import simplejson as json

class MigrationSettings:

    def __init__(self, filename):
        uLogging.info("Read settings from file '%s'." % filename)
        f = open(filename)  # with open(filename) as f  is not supported in python 2.4, Centos5
        json_settings = json.load(f)
        f.close()
        self.filename = filename
        self.plans = self.get_value(json_settings, 'service_plans')
        self.servers = self.get_value(json_settings, 'target_mbx_servers')
        self.concurrency_limit = self.get_value(json_settings, 'concurrency_limit')
        self.failures_limit = self.get_value(json_settings, 'failures_limit')
        self.validate()
        uLogging.debug("Migration is started with parameters \n%s" % json.dumps(json_settings, indent=4))

    def get_value(self, json_settings, parameter_name):
        try:
            return json_settings[parameter_name]
        except KeyError, e:
            raise Exception("Parameter '%s' is missing. Please correct settings in %s." % (parameter_name, self.filename))

    def validate(self):
        if not self.plans:
            raise Exception("Service plans list is empty. Please correct settings in %s." % self.filename)
        for plan in self.plans:
            plan_params = plan.keys()
            if not ('source_plan' in plan_params and 'target_plan' in plan_params and 'subscription_periods' in plan_params):
                raise Exception("Incorrect service plan settings. Source/target plans and subscription periods should be set. Please correct settings in %s." % self.filename)
            if plan['source_plan'] == plan['target_plan']:
                raise Exception("Incorrect service plan settings. Source and target plans cannot be the same. Please correct settings in %s." % self.filename)
            if not plan['subscription_periods']:
                raise Exception("Incorrect service plan settings. Subscription periods mapping is not defined. Please correct settings in %s." % self.filename)

        source_plans = self.get_source_plans()
        unique_source_plans = list(set(source_plans))
        if len(source_plans) != len(unique_source_plans):
            raise Exception("Incorrect service plan settings. The same source plan is present twice. Please correct settings in %s." % self.filename)
        if not self.servers:
            raise Exception("Target mailbox servers list is empty. Please correct settings in %s." % self.filename)
        if self.concurrency_limit <= 0:
            raise Exception("Limit of concurrent flows is too low: it should be positive integer. Please correct settings in %s." % self.filename)

    def get_source_plans(self):
        return [plan['source_plan'] for plan in self.plans]

    def get_source_periods(self, source_plan_id):
        plan_map = self._find_plan_maps_by_source_plan(source_plan_id)[0]
        return [p["source"] for p in plan_map["subscription_periods"]]

    def get_target_plan(self, source_plan_id):
        plan_map = self._find_plan_maps_by_source_plan(source_plan_id)[0]
        return plan_map['target_plan']

    def get_target_period(self, source_plan_id, source_plan_period_id):
        plan_map = self._find_plan_maps_by_source_plan(source_plan_id)[0]
        period_map = self._find_target_period(plan_map, source_plan_period_id)
        return period_map["target"]

    def is_source_plan(self, plan_id):
        plan_maps = self._find_plan_maps_by_source_plan(plan_id)
        return len(plan_maps) > 0

    def is_target_plan(self, plan_id):
        plan_maps = self._find_plan_maps_by_target_plan(plan_id)
        return len(plan_maps) > 0

    def _find_plan_maps_by_source_plan(self, source_plan_id):
        return [p for p in self.plans if p['source_plan'] == source_plan_id]

    def _find_plan_maps_by_target_plan(self, source_plan_id):
        return [p for p in self.plans if p['target_plan'] == source_plan_id]

    @staticmethod
    def _find_target_period(plan_map, source_period_id):
        periods = plan_map["subscription_periods"]
        period_maps = [p for p in periods if p['source'] == source_period_id]
        if not period_maps:
            raise Exception("Could not find target period for source plan %s, source period %s" % (plan_map["source_plan"], source_period_id))
        return [p for p in periods if p['source'] == source_period_id][0]
